package com.flyfish.framework.config;

import com.flyfish.framework.config.properties.SecurityProperties;
import com.flyfish.framework.utils.RSAUtils;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.authentication.AbstractUserDetailsReactiveAuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.userdetails.ReactiveUserDetailsService;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.util.Assert;
import reactor.core.publisher.Mono;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

/**
 * 支持rsa的认证管理器
 *
 * @author wangyu
 */
@Slf4j
public class RSAAuthenticationManager extends AbstractUserDetailsReactiveAuthenticationManager {

    private final ReactiveUserDetailsService userDetailsService;

    private final Boolean rsa;

    public RSAAuthenticationManager(SecurityProperties securityProperties, ReactiveUserDetailsService userDetailsService) {
        Assert.notNull(userDetailsService, "userDetailsService cannot be null");
        this.rsa = securityProperties.isRsa();
        this.userDetailsService = userDetailsService;
    }

    @Override
    public Mono<Authentication> authenticate(Authentication authentication) {
        if (rsa && !authentication.isAuthenticated()) {
            Object credentials = authentication.getCredentials();
            if (credentials instanceof String) {
                Authentication mapped = createAuthentication(authentication);
                return super.authenticate(mapped);
            }
        }
        return super.authenticate(authentication);
    }

    private UsernamePasswordAuthenticationToken createAuthentication(Authentication authentication) throws IllegalArgumentException {
        String password = (String) authentication.getCredentials();
        try {
            password = RSAUtils.decrypt(password, RSAKeys.PRIVATE_KEY);
        } catch (IllegalBlockSizeException | InvalidKeyException | BadPaddingException | NoSuchAlgorithmException | NoSuchPaddingException e) {
            log.error("尝试解密密码出错", e);
            throw new IllegalArgumentException("非法请求！密码格式校验失败！");
        } catch (IllegalArgumentException e) {
            log.error("抛出参数异常", e);
            throw new IllegalArgumentException("密码未加密，请求无效！" + e.getMessage());
        }
        return new UsernamePasswordAuthenticationToken(authentication.getPrincipal(), password, authentication.getAuthorities());
    }

    @Override
    protected Mono<UserDetails> retrieveUser(String username) {
        return this.userDetailsService.findByUsername(username);
    }
}
