package com.flyfish.framework.configuration.resolver;

import com.flyfish.framework.configuration.annotations.PagedQuery;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.utils.UserUtils;
import org.springframework.core.MethodParameter;
import org.springframework.core.ReactiveAdapterRegistry;
import org.springframework.data.domain.Pageable;
import org.springframework.data.util.CastUtils;
import org.springframework.security.core.context.ReactiveSecurityContextHolder;
import org.springframework.web.reactive.BindingContext;
import org.springframework.web.reactive.result.method.HandlerMethodArgumentResolverSupport;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

/**
 * 解析分页请求参数为需要的bean，比map更语义化
 *
 * @author Mr.Wang
 */
public class PageQueryArgumentResolver extends HandlerMethodArgumentResolverSupport {

    /**
     * 这是spirng-data-jpa自带的分页参数解析器，用于在本解析器里解析分页参数
     */
    private ParameterFilter<Pageable> pageableParameterFilter;

    /**
     * 这是spring原生自带的servlet请求数据解析器，用于解析请求参数到bean
     */
    private ModelAttributeFilter modelAttributeFilter;


    public PageQueryArgumentResolver(ReactiveAdapterRegistry adapterRegistry) {
        super(adapterRegistry);
        this.pageableParameterFilter = new PageParameterFilter();
        this.modelAttributeFilter = new ModelAttributeFilter(adapterRegistry);
    }

    @Override
    public boolean supportsParameter(MethodParameter methodParameter) {
        return Qo.class.isAssignableFrom(methodParameter.getParameterType()) &&
                methodParameter.hasParameterAnnotation(PagedQuery.class);
    }

    /**
     * 解析结果
     *
     * @param parameter      方法参数
     * @param bindingContext 上下文
     * @param exchange       交换数据
     * @return 结果
     */
    @Override
    public Mono<Object> resolveArgument(MethodParameter parameter, BindingContext bindingContext, ServerWebExchange exchange) {
        // 使用spring的hack来解析pageable
        Pageable pageable = pageableParameterFilter.filter(parameter, bindingContext, exchange);
        // 使用spring的默认解析器解析所有变量
        return modelAttributeFilter.filter(parameter, bindingContext, exchange).flatMap(qo -> {
            if (qo instanceof Qo) {
                Qo<?> query = (Qo<?>) qo;
                query.setPageable(pageable);
                return ReactiveSecurityContextHolder.getContext()
                        .map(UserUtils::extractUserDetails)
                        .map(user -> {
                            query.setUser(user);
                            return query;
                        }).defaultIfEmpty(CastUtils.cast(query));
            }
            return Mono.just(qo);
        });
    }
}
