package com.flyfish.framework.service.impl;

import com.flyfish.framework.auditor.ReactiveBeanAuditor;
import com.flyfish.framework.auditor.ReactiveBeanPoster;
import com.flyfish.framework.bean.SyncVo;
import com.flyfish.framework.domain.authorized.AuthorizedDomain;
import com.flyfish.framework.domain.base.AuditDomain;
import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.exception.biz.InvalidBusinessException;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import com.flyfish.framework.service.BaseReactiveService;
import com.flyfish.framework.utils.Assert;
import com.flyfish.framework.utils.CopyUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Example;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.mongodb.repository.query.MongoEntityInformation;
import org.springframework.data.util.CastUtils;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * 基础的Service，继承常用的全部Crud操作，基于JPARepository
 * 已完全实现全部webflux功能，提供全异步支持，系统性能最大化
 *
 * @param <T> 实体泛型
 */
public class BaseReactiveServiceImpl<T extends Domain> implements BaseReactiveService<T> {

    @Autowired
    protected DefaultReactiveRepository<T> repository;
    @Autowired(required = false)
    protected ReactiveBeanAuditor<T> auditor;
    @Autowired(required = false)
    protected ReactiveBeanPoster<T> poster;
    @Autowired
    private ReactiveBeanAuditor<Domain> commonAuditor;
    @Autowired
    private ReactiveBeanAuditor<AuditDomain> operationAuditor;
    @Autowired
    private ReactiveBeanAuditor<AuthorizedDomain> authorizeAuditor;

    /**
     * 查询
     *
     * @param entity 实体本身
     * @return 结果
     */
    @Override
    public Mono<T> getOne(T entity) {
        return repository.findOne(Example.of(entity));
    }

    /**
     * 查询
     *
     * @param query 查询模型
     * @return 结果
     */
    @Override
    public Mono<T> getOne(Qo<T> query) {
        return repository.findOne(query);
    }

    /**
     * 通过Id查询
     *
     * @param id 主键
     * @return 结果
     */
    @Override
    public Mono<T> getById(String id) {
        return repository.findById(id);
    }

    /**
     * 通过实体名称访问
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Mono<T> getByName(String name) {
        return repository.findByName(name);
    }

    /**
     * 根据ID集合来查询
     *
     * @param ids id列表
     * @return 结果
     */
    @Override
    public Flux<T> getByIds(List<String> ids) {
        return repository.findAllById(ids);
    }

    /**
     * 通过某个字段的集合包含查询
     *
     * @param key    键
     * @param values 值
     * @return 结果
     */
    @Override
    public <K> Flux<T> getByValues(String key, List<K> values) {
        return repository.findAllByValues(key, values);
    }

    /**
     * 查询列表
     *
     * @param entity 实体本身
     * @return 结果
     */
    @Override
    public Flux<T> getList(T entity) {
        return repository.findAll(Example.of(entity));
    }

    /**
     * 查询列表
     *
     * @param query 查询
     * @return 结果
     */
    @Override
    public Flux<T> getList(Qo<T> query) {
        return repository.findAll(query);
    }

    /**
     * 分页查询列表
     *
     * @param entity   实体本身
     * @param pageable 分页
     * @return 结果
     */
    @Override
    public Mono<Page<T>> getPageList(T entity, Pageable pageable) {
        return repository.findAll(Example.of(entity), pageable.getSort())
                .buffer(pageable.getPageSize(), (int) pageable.getOffset())
                .reduce(new PageImpl<>(new ArrayList<>()), (result, item) -> {
                    List<T> list = result.getContent();
                    list.addAll(item);
                    return new PageImpl<>(list, pageable, result.getTotalElements());
                });
    }

    /**
     * 分页查询列表
     *
     * @param query 查询实体
     * @return 结果
     */
    @Override
    public Mono<Page<T>> getPageList(Qo<T> query) {
        return repository.findAll(query, query.getPageable());
    }

    /**
     * 获取所有对象
     *
     * @return 结果
     */
    @Override
    public Flux<T> getAll() {
        return repository.findAll();
    }

    /**
     * 查询总记录数
     *
     * @return 总数
     */
    @Override
    public Mono<Long> countAll() {
        return repository.count();
    }

    /**
     * 查询总记录数
     *
     * @param entity 通过实体查询个数
     * @return 结果
     */
    @Override
    public Mono<Long> count(T entity) {
        return repository.count(Example.of(entity));
    }

    /**
     * 查询总记录数
     *
     * @param query 查询实体
     * @return 结果
     */
    @Override
    public Mono<Long> count(Qo<T> query) {
        return repository.count(query);
    }

    /**
     * 查询是否存在
     *
     * @param query 查询
     * @return 结果
     */
    @Override
    public Mono<Boolean> exists(Qo<T> query) {
        return repository.exists(query);
    }

    /**
     * 插入新记录
     *
     * @param entity 实体
     */
    @Override
    public Mono<T> create(T entity) {
        return audit(entity).flatMap(repository::insert).flatMap(this::post);
    }

    /**
     * 插入 不插入null字段
     *
     * @param entity 实体
     */
    @Override
    public Mono<T> createSelective(T entity) {
        return audit(entity).flatMap(repository::insert).flatMap(this::post);
    }

    /**
     * 删除
     *
     * @param entity 实体
     */
    @Override
    public Mono<Void> delete(T entity) {
        return repository.delete(entity);
    }

    /**
     * 根据Id删除
     *
     * @param id 主键
     */
    @Override
    public Mono<Void> deleteById(String id) {
        return repository.deleteById(id);
    }

    /**
     * 根据id更新
     *
     * @param entity 实体
     */
    @Override
    public Mono<T> updateById(T entity) {
        return audit(entity).flatMap(repository::save).flatMap(this::post);
    }

    /**
     * 不update null
     *
     * @param entity 实体
     */
    @Override
    public Mono<T> updateSelectiveById(T entity) {
        Assert.hasText(entity.getId(), "更新的主键不可为空！");
        return repository.findById(entity.getId())
                .map(saved -> CopyUtils.copyProps(entity, saved))
                .flatMap(this::updateById)
                .switchIfEmpty(Mono.defer(() -> Mono.error(new InvalidBusinessException("要更新的信息不存在"))));
    }

    /**
     * 根据ID集合批量删除
     *
     * @param ids id结合
     */
    @Override
    public Mono<Void> deleteBatchByIds(List<String> ids) {
        return Flux.fromIterable(ids)
                .flatMap(t -> repository.deleteById(t))
                .then(Mono.empty());
    }

    /**
     * 通过某一个键删除所有
     *
     * @param qo 查询实体
     */
    @Override
    public Mono<Void> deleteAll(Qo<T> qo) {
        return repository.deleteAll(qo);
    }

    /**
     * 通过某一个键删除所有
     *
     * @param entity 查询
     */
    @Override
    public Mono<Void> deleteAll(T entity) {
        return repository.delete(entity);
    }

    /**
     * 删除全部实体，危险！
     */
    @Override
    public Mono<Void> deleteAll() {
        return repository.deleteAll();
    }

    /**
     * 批量更新
     *
     * @param entities 要批量更新的集合
     */
    @Override
    public Flux<T> updateBatch(List<T> entities) {
        if (CollectionUtils.isNotEmpty(entities)) {
            // 带有id的集合
            List<String> ids = entities.stream()
                    .filter(Objects::nonNull)
                    .map(Domain::getId)
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
            return Mono.just(ids)
                    .filter(CollectionUtils::isNotEmpty)
                    .flatMapMany(list -> repository.findAllById(list))
                    .collectMap(Domain::getId, t -> t)
                    .flatMap(map -> Flux.fromIterable(entities)
                            // 补全已保存信息
                            .map(t -> map.containsKey(t.getId()) ? CopyUtils.copyProps(t, map.get(t.getId())) : t)
                            .flatMap(this::audit)
                            .collectList())
                    .flatMapMany(list -> repository.saveAll(list))
                    .flatMap(this::post);
        }
        return Flux.fromIterable(Collections.emptyList());
    }

    /**
     * 同步数据
     *
     * @param entities 数据们
     * @return 结果
     */
    @Override
    public Mono<SyncVo<T>> sync(List<T> entities) {
        return this.updateBatch(entities)
                .collectList()
                .map(list -> SyncVo.<T>builder()
                        .success(list.size())
                        .list(list)
                        .build());
    }

    /**
     * 对象审查
     *
     * @param entity 实体
     */
    @SuppressWarnings("unchecked")
    protected Mono<T> audit(T entity) {
        Mono<T> mono = Mono.just(entity);
        // 首先自定义审查，可能返回自定义对象
        if (auditor != null) {
            mono = mono.flatMap(auditor::audit);
        }
        // 然后公共审查
        mono = (Mono<T>) mono.flatMap(commonAuditor::audit);
        // 接着操作信息审查
        if (entity instanceof AuditDomain) {
            mono = (Mono<T>) mono.cast(AuditDomain.class).flatMap(operationAuditor::audit);
        }
        // 最后权限审查
        if (entity instanceof AuthorizedDomain) {
            mono = (Mono<T>) mono.cast(AuthorizedDomain.class).flatMap(authorizeAuditor::audit);
        }
        return mono;
    }

    /**
     * 后置审查
     *
     * @param entity 实体
     * @return 结果
     */
    protected Mono<T> post(T entity) {
        if (null != poster) {
            return poster.post(entity);
        }
        return Mono.just(entity);
    }

    public <R extends DefaultReactiveRepository<T>> R getRepository() {
        return CastUtils.cast(repository);
    }

    /**
     * 获取集合名称
     *
     * @return 结果
     */
    @Override
    public MongoEntityInformation<? extends Domain, String> getEntityInformation() {
        return repository.getEntityInformation();
    }
}

